/****************************************************************************
 *~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 *~                             IR Kernel                                    ~
 *~                         Copyright(C) 1993                                ~
 *~                       Software Kinetics Ltd.                             ~
 *~                        All Rights Reserved                               ~
 *~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 * Copyright Software Kinetics Ltd.
 * 
 * Software Kinetics Ltd. has granted a licence of this software to
 * National Library of Canada.
 *
 * The National Library of Canada grants to the user a royalty-free
 * non-exclusive licence in perpetuity to use, modify, improve,
 * reproduce and sell the software.
 *
 * The software is provided to the user "as is" without warranty of
 * any kind.  Neither Software Kinetics Ltd. nor the National Library
 * of Canada warrants,guarantees or makes any representation regarding
 * the use, or the results of the use of the software in terms of
 * accuracy,reliability, or otherwise.  The entire risk as results and
 * performance of the software is assumed by the user. Neither
 * Software Kinetics Ltd. nor the National Library of Canada shall be
 * liable for any direct, indirect, consequential or incidental
 * damages arising out the use of software.
 *~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ 
 *
 * Program: ircksrch.c
 *
 * Purpose: This file contains the routines to check the mandatory fields of
 *          search request PDU and search response PDU.
 *
 * Invocation:
 *
 * Input Arguments:
 *
 * Output/Exit Values: None
 *
 * Functions used:
 *
 * Version: Date:   94/05/24
 *          Name:   Benjamin Yen
 *          Reason: First Draft of IR Target Kernel
 *
 * Reference: IR KERNEL - TARGET PRELIMINARY SOFTWARE DESIGN DOCUMENT
 *            IR KERNEL - TARGET SOFTWARE DETAIL DESIGN DOCUMENT
 *            IR KERNEL - INTERFACE REQUIREMENTS SPECIFICATION
 *            OSI - ANSI Z39.50 DOCUMENTATION
 *
 * Notes:
 *
 * $Id: $
 * $Log: $
 *
 *
 *****************************************************************************/

#include <stdio.h>
#include <rpc/rpc.h>
#include <fcntl.h>
#include <pwd.h>
#include <ctype.h>
#include <sys/errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <arpa/inet.h>
#include "ir.h"
#include "ir_oid.h"
#include "asn_incl.h"
#include "irerror.h"
#include "irconst.h"
#include "oid_tool.c"

/**************************** Extern Global Ids **************************/

extern int errno;

/*****************************************************************************
 *
 * Function: ir_fill_search_response
 *
 * Purpose: This function fills and creates the search response PDU with error 
 *          code 8 indicating invalid mandatory fields.
 *
 * Input Arguments:
 *              PDU_sreq - search request PDU pointer
 *
 * Return values: SUCCESS or FAILED
 *
 * Output Arguments:
 *              PDU_sresp - search response PDU pointer
 *
 * External Variables:
 *
 * Function used:
 *
 * Algorithm:
 *
 * Version: Date:   94/05/24
 *          Name:   Benjamin Yen
 *          Reason: First Draft of IR Target Kernel
 *
 * Reference:
 *
 * Notes:
 *
 ****************************************************************************/
ir_fill_search_response(PDU_sreq, PDU_sresp)
    APDU *PDU_sreq;
    APDU *PDU_sresp;
{
    char error_str[73];

#ifdef DEBUG_LEVEL1
    fprintf(DEBUG_FILE1, "Calling ir_fill_search_response\n");
#endif

    PDU_sresp->a.searchResp = (struct SearchResp *) Asn1Alloc(sizeof(struct SearchResp));

    PDU_sresp->choiceId = APDU_SEARCHRESP;

    PDU_sresp->a.searchResp->referenceId.octetLen = PDU_sreq->a.searchReq->referenceId.octetLen;
    PDU_sresp->a.searchResp->referenceId.octs = (char *) Asn1Alloc(PDU_sreq->a.searchReq->referenceId.octetLen);
    memcpy(PDU_sresp->a.searchResp->referenceId.octs, PDU_sreq->a.searchReq->referenceId.octs, PDU_sreq->a.searchReq->referenceId.octetLen);

    PDU_sresp->a.searchResp->resultCount = 1L;
    PDU_sresp->a.searchResp->numRecordsReturned = 1L;

    PDU_sresp->a.searchResp->nextResultSetPosition = 0L;
    PDU_sresp->a.searchResp->searchStatus = 1L;

    PDU_sresp->a.searchResp->resultSetStatus = (SearchRespInt *) Asn1Alloc(sizeof(SearchRespInt));
    PDU_sresp->a.searchResp->presentStatus = (PresentStatus *) Asn1Alloc(sizeof(PresentStatus));
    *PDU_sresp->a.searchResp->resultSetStatus = 0;
    *PDU_sresp->a.searchResp->presentStatus = 0;

    PDU_sresp->a.searchResp->dbOrDiagnosticRec = (struct Records *) Asn1Alloc(sizeof(struct Records));

    /* create diagnostics record (error record) */
	PDU_sresp->a.searchResp->dbOrDiagnosticRec->a.nonSurDiagnostics = (struct DiagRec *) Asn1Alloc(sizeof(struct DiagRec));

    PDU_sresp->a.searchResp->dbOrDiagnosticRec->choiceId = RECORDS_NONSURDIAGNOSTICS;

	ASN_MakeOid((char *) OID_DIAGNOSTIC_SET_BIB1, &PDU_sresp->a.searchResp->dbOrDiagnosticRec->a.nonSurDiagnostics->diagnosticSetId);

	PDU_sresp->a.searchResp->dbOrDiagnosticRec->a.nonSurDiagnostics->condition = 8; /* error code */
	strcpy(error_str, "THE MANDATORY FIELDS ARE INVALID");

	PDU_sresp->a.searchResp->dbOrDiagnosticRec->a.nonSurDiagnostics->addInfo.octetLen = strlen(error_str);
	PDU_sresp->a.searchResp->dbOrDiagnosticRec->a.nonSurDiagnostics->addInfo.octs = (char *) Asn1Alloc(PDU_sresp->a.searchResp->dbOrDiagnosticRec->a.nonSurDiagnostics->addInfo.octetLen);
	memcpy(PDU_sresp->a.searchResp->dbOrDiagnosticRec->a.nonSurDiagnostics->addInfo.octs, error_str, PDU_sresp->a.searchResp->dbOrDiagnosticRec->a.nonSurDiagnostics->addInfo.octetLen);

#ifdef DEBUG_LEVEL1
    fprintf(DEBUG_FILE1, "Exiting ir_fill_search_response\n");
#endif

    return (SUCCESS);
}

/*****************************************************************************
 *
 * Function: ir_create_search_response
 *
 * Purpose: This function creates the search response PDU, encodes, and
 *          search response APDU for sending to client.
 *
 * Input Arguments:
 *          PDU_sreq - search request PDU pointer
 *          data_ptr - temporary data buffer for encoding process
 *
 * Return values: SUCCESS or FAILED
 *
 * Output Arguments:
 *              outputBuf - encoded APDU pointer
 *              APDU_length - the size of the encoded APDU
 *
 * External Variables:
 *
 * Function used:
 *              ir_fill_search_response()
 *              SBufInstallData()
 *              SBufResetInWriteRvsMode()
 *              BEncAPDU()
 *              ir_fwrite()
 *
 * Algorithm:
 *
 * Version: Date:   94/05/24
 *          Name:   Benjamin Yen
 *          Reason: First Draft of IR Target Kernel
 *
 * Reference:
 *
 * Notes:
 *
 ****************************************************************************/
ir_create_search_response(PDU_sreq, data_ptr, outputBuf, APDU_length)
    APDU *PDU_sreq;
    char **data_ptr;
    SBuf *outputBuf;
    int *APDU_length;
{
    APDU PDU_ptr;

#ifdef DEBUG_LEVEL1
    fprintf(DEBUG_FILE1, "Calling ir_create_search_response()\n");
#endif

    ir_fill_search_response(PDU_sreq, &PDU_ptr);

#ifdef DEBUG_LEVEL2
    fprintf(DEBUG_FILE2, "ir_create_search_response: begin dump data:\n");
    PrintSearchResp(DEBUG_FILE2, PDU_ptr.a.searchResp, 0);
    fprintf(DEBUG_FILE2, "ir_create_search_response: end dump data:\n");
#endif

    SBufInstallData(outputBuf, (*data_ptr), MAXMESSAGESIZE * 2);
    SBufResetInWriteRvsMode(outputBuf);

    *APDU_length = BEncAPDU(outputBuf, &PDU_ptr);

	if (SBufWriteError(outputBuf))
	{
		display_error("KERNEL", ERR_ENCODE_APDU, -1);
		return(FAILED);
	}

#ifdef DEBUG_LEVEL2
    fprintf(DEBUG_FILE2, "ir_create_search_response: SEARCH Response APDU Dump\n");
    ir_fwrite(SBufDataPtr(outputBuf), SBufDataLen(outputBuf), 1, DEBUG_FILE2);
#endif

    return (SUCCESS);
}

/*****************************************************************************
 *
 * Function: ir_actual_check_search_request
 *
 * Purpose: This function checks the mandatory fields of search request PDU 
 *          for validity.
 *
 * Input Arguments: PDU_ptr - search request PDU pointer
 *
 * Return values: SUCCESS or FAILED
 *
 * Output Arguments:
 *
 * External Variables:
 *
 * Function used:
 *              display_error()
 *
 * Algorithm:
 *
 * Version: Date:   94/05/24
 *          Name:   Benjamin Yen
 *          Reason: First Draft of IR Target Kernel
 *
 * Reference:
 *
 * Notes:
 *
 ****************************************************************************/
ir_actual_check_search_request(PDU_ptr)
    APDU *PDU_ptr;
{

    if (PDU_ptr->a.searchReq->smallSetUpperBound == 0 &&
            PDU_ptr->a.searchReq->largeSetLowerBound == 1)
    {
        /*
         * The client system doesn't want to include the result in the
         * response
         */
    }
    else
    {
        if (PDU_ptr->a.searchReq->smallSetUpperBound <= 0)
        {
            display_error("KERNEL", ERR_SEARCH_SMALL_SET_IS_INVALID, -1);
            return (FAILED);
        }

        if (PDU_ptr->a.searchReq->largeSetLowerBound <= 0)
        {
            display_error("KERNEL", ERR_SEARCH_LARGE_SET_IS_INVALID, -1);
            return (FAILED);
        }

        if ((PDU_ptr->a.searchReq->smallSetUpperBound >= PDU_ptr->a.searchReq->largeSetLowerBound) ||
                (PDU_ptr->a.searchReq->smallSetUpperBound > PDU_ptr->a.searchReq->mediumSetPresentNum))
        {
            display_error("KERNEL", ERR_SEARCH_SMALL_SET_IS_TOO_LARGE, -1);
            return (FAILED);
        }

        if (PDU_ptr->a.searchReq->mediumSetPresentNum >= PDU_ptr->a.searchReq->largeSetLowerBound)
        {
            display_error("KERNEL", ERR_SEARCH_MEDIUM_SET_IS_TOO_LARGE, -1);
            return (FAILED);
        }
    }

    if (!(PDU_ptr->a.searchReq->replaceIndicator == 0 || PDU_ptr->a.searchReq->replaceIndicator == 1))
    {
        display_error("KERNEL", ERR_INVALID_SEARCH_REPLACE_INDICATOR, -1);
        return (FAILED);
    }

    if (PDU_ptr->a.searchReq->resultSetName.octetLen <= 0 || PDU_ptr->a.searchReq->resultSetName.octs == NULL)
    {
        display_error("KERNEL", ERR_EMPTY_SEARCH_RESULT_SET_NAME, -1);
        return (FAILED);
    }

    if ((PDU_ptr->a.searchReq->databaseNames->count <= 0) ||
            (PDU_ptr->a.searchReq->databaseNames->dataSize <= 0) ||
            (PDU_ptr->a.searchReq->databaseNames->first == NULL)
        )
    {
        display_error("KERNEL", ERR_EMPTY_SEARCH_DATABASE_NAME, -1);
        return (FAILED);
    }
    else if (PDU_ptr->a.searchReq->databaseNames->first->data == NULL)
    {
        display_error("KERNEL", ERR_EMPTY_SEARCH_DATABASE_NAME, -1);
        return (FAILED);
    }

    if (!(PDU_ptr->a.searchReq->query->choiceId == QUERY_TYPE_0 ||
          PDU_ptr->a.searchReq->query->choiceId == QUERY_TYPE_1 ||
          PDU_ptr->a.searchReq->query->choiceId == QUERY_TYPE_2 ||
          PDU_ptr->a.searchReq->query->choiceId == QUERY_TYPE_100 ||
          PDU_ptr->a.searchReq->query->choiceId == QUERY_TYPE_101
          )
        )
    {
        display_error("KERNEL", ERR_INVALID_SEARCH_QUERY_TYPE, -1);
        return (FAILED);
    }

    switch (PDU_ptr->a.searchReq->query->choiceId)
    {
        case QUERY_TYPE_0:
            if (PDU_ptr->a.searchReq->query->a.type_0->octetLen <= 0 ||
                    PDU_ptr->a.searchReq->query->a.type_0->octs == NULL)
            {
                display_error("KERNEL", ERR_EMPTY_SEARCH_QUERY, -1);
                return (FAILED);
            }
            break;
        case QUERY_TYPE_1:
            if (PDU_ptr->a.searchReq->query->a.type_1->attributeSetId.octetLen <= 0 ||
                    PDU_ptr->a.searchReq->query->a.type_1->attributeSetId.octs == NULL ||
                PDU_ptr->a.searchReq->query->a.type_1->rPNStructure == NULL)
            {
                display_error("KERNEL", ERR_EMPTY_SEARCH_QUERY, -1);
                return (FAILED);
            }
            break;
        case QUERY_TYPE_2:
            if (PDU_ptr->a.searchReq->query->a.type_2->octetLen <= 0 ||
                    PDU_ptr->a.searchReq->query->a.type_2->octs == NULL)
            {
                display_error("KERNEL", ERR_EMPTY_SEARCH_QUERY, -1);
                return (FAILED);
            }
            break;
        case QUERY_TYPE_100:
            if (PDU_ptr->a.searchReq->query->a.type_100->octetLen <= 0 ||
                    PDU_ptr->a.searchReq->query->a.type_100->octs == NULL)
            {
                display_error("KERNEL", ERR_EMPTY_SEARCH_QUERY, -1);
                return (FAILED);
            }
            break;
        case QUERY_TYPE_101:
            if (PDU_ptr->a.searchReq->query->a.type_101->octetLen <= 0 ||
                    PDU_ptr->a.searchReq->query->a.type_101->octs == NULL)
            {
                display_error("KERNEL", ERR_EMPTY_SEARCH_QUERY, -1);
                return (FAILED);
            }
            break;
    }

    return (SUCCESS);
}

/*****************************************************************************
 *
 * Function: ir_check_search_request
 *
 * Purpose: This function performs the following functions:
 *              - checks the mandatory fields of search request PDU,
 *              - creates and sends the search response APDU to client if
 *                the search request PDU has any invalid fields.
 *
 * Input Arguments:
 *              PDU - search request PDU pointer
 *              data_ptr - temporary data pointer for ASN functions
 *              socketFd - the socket Fd for communicating to IR kernel
 *
 * Return values: SUCCESS or FAILED
 *
 * Output Arguments:
 *
 * External Variables:
 *
 * Function used:
 *              ir_check_search_request()
 *              ir_create_search_response()
 *              tpa_send_APDU()
 *              display_error()
 *
 * Algorithm:
 *
 * Version: Date:   94/05/24
 *          Name:   Benjamin Yen
 *          Reason: First Draft of IR Target Kernel
 *
 * Reference:
 *
 * Notes:
 *
 ****************************************************************************/
ir_check_search_request(PDU_sreq, data_ptr, socketFd)
    APDU *PDU_sreq;
    char **data_ptr;
    int socketFd;
{
    int result;
    long APDU_length;
    SBuf APDU_sresp;

#ifdef DEBUG_LEVEL1
    fprintf(DEBUG_FILE1, "KERNEL: calling ir_check_search_request\n");
#endif

    result = ir_actual_check_search_request(PDU_sreq);

    if (result == FAILED)
    {
        /* create the diagnostic record due to incomplete mandatory fields */
        ir_create_search_response(PDU_sreq, data_ptr, &APDU_sresp, &APDU_length);

#ifdef DEBUG_LEVEL3
			fprintf(DEBUG_FILE3, "ir_check_search_request: %ld search response APDU length\n", APDU_length);
#endif

		if (cs_send_data(APDU_sresp.dataStart, (int) APDU_length, socketFd) == FAILED)
		{
			display_error("KERNEL", ERR_SEND_DATA_TO_CLIENT, errno);
			return (FAILED);
		}
	}

    return (SUCCESS);
}

/*****************************************************************************
 *
 * Function: ir_check_search_response
 *
 * Purpose: This function checks the mandatory fields of search response PDU 
 *          for validity.
 *
 * Input Arguments: PDU_ptr - search response PDU pointer
 *
 * Return values: SUCCESS or FAILED
 *
 * Output Arguments:
 *
 * External Variables:
 *
 * Function used:
 *              display_error()
 *
 * Algorithm:
 *
 * Version: Date:   94/05/24
 *          Name:   Benjamin Yen
 *          Reason: First Draft of IR Target Kernel
 *
 * Reference:
 *
 * Notes:
 *
 ****************************************************************************/
ir_check_search_response(PDU_ptr)
    APDU *PDU_ptr;
{

    if (PDU_ptr->a.searchResp->resultCount < 0)
    {
        display_error("KERNEL", ERR_EMPTY_SEARCH_RESP_RESULT_COUNT, -1);
        return (FAILED);
    }

    if (PDU_ptr->a.searchResp->numRecordsReturned < 0)
    {
        display_error("KERNEL", ERR_EMPTY_SEARCH_RESP_NO_OF_RECS_RETURN, -1);
        return (FAILED);
    }

    if (PDU_ptr->a.searchResp->nextResultSetPosition < 0)
    {
        display_error("KERNEL", ERR_EMPTY_SEARCH_RESP_NEXT_RESULT_SET_POSITION, -1);
        return (FAILED);
    }

    if (PDU_ptr->a.searchResp->searchStatus != TRUE && PDU_ptr->a.searchResp->searchStatus != FALSE)
    {
        display_error("KERNEL", ERR_EMPTY_SEARCH_RESP_SEARCH_STATUS, -1);
        return (FAILED);
    }

    return (SUCCESS);
}
